@@ -1,6 +1,7 @@
 #define PERL_NO_GET_CONTEXT
 #include "EXTERN.h"
 #include "perl.h"
+#define NO_XSLOCKS
 #include "XSUB.h"
 
 #include "message.h"
@@ -83,9 +84,10 @@ void store_self(pTHX, mthread* thread) {
 mthread* S_get_self(pTHX) {
 	SV** self_sv = hv_fetch(PL_modglobal, "threads::lite::thread", 21, FALSE);
 	if (!self_sv) {
+		mthread* ret;
 		if (ckWARN(WARN_THREADS))
 			Perl_warn(aTHX, "Creating thread context where non existed\n");
-		mthread* ret = mthread_alloc(aTHX);
+		ret = mthread_alloc(aTHX);
 		store_self(aTHX, ret);
 		return ret;
 	}
@@ -105,21 +107,23 @@ perl_mutex* get_shutdown_mutex() {
 static void* run_thread(void* arg) {
 	mthread* thread = (mthread*) arg;
 	PerlInterpreter* my_perl = thread->interp;
+	message to_run, message;
+	SV *call, *status;
+	perl_mutex* shutdown_mutex;
+
+	dSP;
 
 #ifndef WIN32
 	S_set_sigmask(&thread->initial_sigmask);
 #endif
 	PERL_SET_CONTEXT(my_perl);
 
-	message to_run;
 	queue_dequeue(&thread->queue, &to_run, NULL);
-	SV* call = SvRV(message_load_value(&to_run));
-
-	dSP;
+	call = SvRV(message_load_value(&to_run));
 
 	PUSHMARK(SP);
 	PUSHs(newSVpvn("exit", 4));
-	SV* status = newSVpvn("normal", 6);
+	status = newSVpvn("normal", 6);
 	PUSHs(status);
 	PUSHs(newSViv(thread->id));
 
@@ -133,20 +137,21 @@ static void* run_thread(void* arg) {
 		warn("Thread %"UVuf" got error %s\n", thread->id, SvPV_nolen(ERRSV));
 		PUSHs(ERRSV);
 	}
-	message message;
 	message_from_stack_pushed(&message);
 	send_listeners(thread, &message);
 	message_destroy(&message);
 
-	perl_mutex* shutdown_mutex = get_shutdown_mutex();
+	shutdown_mutex = get_shutdown_mutex();
+
 	MUTEX_LOCK(shutdown_mutex);
 	perl_destruct(my_perl);
 	MUTEX_UNLOCK(shutdown_mutex);
 
 	mthread_destroy(thread);
+
+	PerlMemShared_free(thread);
+
 	perl_free(my_perl);
-	
-	Safefree(thread);
 
 	return NULL;
 }
@@ -194,14 +199,16 @@ static mthread* start_thread(mthread* thread, IV stack_size) {
 #ifdef WIN32
 	CreateThread(NULL, (DWORD)stack_size, run_thread, (LPVOID)thread, STACK_SIZE_PARAM_IS_A_RESERVATION, NULL);
 #else
-	int rc_stack_size = 0;
-	int rc_thread_create = 0;
-
-	S_block_most_signals(&thread->initial_sigmask);
-
 	static pthread_attr_t attr;
 	static int attr_inited = 0;
 	static int attr_detached = PTHREAD_CREATE_DETACHED;
+
+	int rc_stack_size = 0;
+	int rc_thread_create = 0;
+	pthread_t thr;
+
+	S_block_most_signals(&thread->initial_sigmask);
+
 	if (! attr_inited) {
 		pthread_attr_init(&attr);
 		attr_inited = 1;
@@ -214,12 +221,10 @@ static mthread* start_thread(mthread* thread, IV stack_size) {
 
 #  ifdef _POSIX_THREAD_ATTR_STACKSIZE
 	/* Set thread's stack size */
-	if (stack_size > 0) {
+	if (stack_size > 0)
 		rc_stack_size = pthread_attr_setstacksize(&attr, (size_t)stack_size);
-	}
 #  endif
 
-	pthread_t thr;
 	/* Create the thread */
 	if (! rc_stack_size) {
 #  ifdef OLD_PTHREADS_API
@@ -252,24 +257,18 @@ static PerlInterpreter* construct_perl() {
 	return my_perl;
 }
 
-static int get_clone_number(pTHX, HV* options) {
-	SV** clone_number_ptr = hv_fetch(options, "pool_size", 9, FALSE);
-	if (clone_number_ptr && SvOK(*clone_number_ptr))
-		return SvIV(*clone_number_ptr);
-	return 1;
-}
-
-static int should_monitor(pTHX, HV* options) {
-	SV** monitor_ptr = hv_fetch(options, "monitor", 7, FALSE);
-	if (monitor_ptr && SvOK(*monitor_ptr))
-		return SvIV(*monitor_ptr);
-	return FALSE;
-}
-
-static void load_modules(pTHX, HV* options) {
+static void save_modules(pTHX, message* message, HV* options) {
 	SV** modules_ptr = hv_fetch(options, "modules", 7, FALSE);
-	if (modules_ptr && SvROK(*modules_ptr) && SvTYPE(SvRV(*modules_ptr)) == SVt_PVAV) {
-		AV* list = (AV*)SvRV(*modules_ptr);
+	if (modules_ptr && SvROK(*modules_ptr) && SvTYPE(SvRV(*modules_ptr)) == SVt_PVAV)
+		message_store_value(message, SvRV(*modules_ptr));
+	else
+		Zero(message, 1, message);
+}
+
+static void load_modules(pTHX, message* list_mess) {
+	if (list_mess->type) {
+		SV* list_ref = message_load_value(list_mess);
+		AV* list = (AV*)SvRV(list_ref);
 		I32 len = av_len(list) + 1;
 		int i;
 		for(i = 0; i < len; i++) {
@@ -279,13 +278,6 @@ static void load_modules(pTHX, HV* options) {
 	}
 }
 
-static unsigned get_stack_size(pTHX, HV* options) {
-	SV** stack_size_ptr = hv_fetch(options, "stack_size", 10, FALSE);
-	if (stack_size_ptr && SvOK(*stack_size_ptr))
-		return SvUV(*stack_size_ptr);
-	return 65536u;
-}
-
 static void push_thread(pTHX, mthread* thread) {
 	PERL_SET_CONTEXT(aTHX);
 	dSP;
@@ -295,42 +287,77 @@ static void push_thread(pTHX, mthread* thread) {
 	PUTBACK;
 }
 
-void S_create_push_threads(PerlInterpreter* self, HV* options, SV* startup) {
-	UV id = S_get_self(self)->id;
+struct thread_create {
+	UV parent_id;
 	message to_run;
-	S_message_store_value(self, &to_run, startup);
-	SvREFCNT_inc_NN(options);
+	message modules;
+	int monitor;
+	size_t stack_size;
+};
 
-	int clone_number = get_clone_number(self, options);
-	int monitor = should_monitor(self, options);
-	size_t stack_size = get_stack_size(self, options);
+static IV get_iv_option(pTHX_ HV* options, const char* key, IV default_value) {
+	SV** value__ptr = hv_fetch(options, key, strlen(key), FALSE);
+	if (value__ptr && SvOK(*value__ptr))
+		return SvIV(*value__ptr);
+	return default_value;
+}
+
+static int prepare_thread_create(pTHX, struct thread_create* new_thread, HV* options, SV* startup) {
+	UV id = get_self()->id;
+
+	message_store_value(&new_thread->to_run, startup);
+
+	save_modules(aTHX, &new_thread->modules, options);
+
+	new_thread->monitor = get_iv_option(aTHX, options, "monitor", FALSE);
+	new_thread->stack_size = get_iv_option(aTHX, options, "stack_size", 65536);
+	return get_iv_option(aTHX, options, "pool_size", 1);
+}
+
+static PerlInterpreter* thread_clone(pTHX, mthread* thread) {
+	dXCPT;
+	XCPT_TRY_START {
+		return perl_clone(my_perl, 0);
+	} XCPT_TRY_END;
+	XCPT_CATCH {
+		Perl_warn(aTHX_ "Cought exception, rethrowing\n");
+		mthread_destroy(thread);
+		XCPT_RETHROW;
+	}
+}
+
+void S_create_push_threads(PerlInterpreter* self, HV* options, SV* startup) {
+	struct thread_create thread_options;
+
+	Zero(&thread_options, 1, struct thread_create);
+	int clone_number = prepare_thread_create(self, &thread_options, options, startup);
 
 	PerlInterpreter* my_perl = construct_perl();
-	load_modules(my_perl, options);
+	load_modules(my_perl, &thread_options.modules);
 
 	mthread* thread = mthread_alloc(my_perl);
 	store_self(my_perl, thread);
-	if (monitor)
-		thread_add_listener(self, thread->id, id);
-
-	queue_enqueue(&thread->queue, &to_run, NULL);
+	if (thread_options.monitor)
+		thread_add_listener(self, thread->id, thread_options.parent_id);
 
 	push_thread(self, thread);
 	while (--clone_number) {
-		PerlInterpreter* my_clone = perl_clone(my_perl, 0);
+		PerlInterpreter* my_clone = thread_clone(my_perl, thread);
 		mthread* thread = mthread_alloc(my_clone);
 		message clone;
 
 		store_self(my_clone, thread);
-		if (monitor)
-			thread_add_listener(self, thread->id, id);
-		message_clone(&to_run, &clone);
+		if (thread_options.monitor)
+			thread_add_listener(self, thread->id, thread_options.parent_id);
+		message_clone(&thread_options.to_run, &clone);
 		queue_enqueue(&thread->queue, &clone, NULL);
 		push_thread(self, thread);
-		start_thread(thread, stack_size);
+		start_thread(thread, thread_options.stack_size);
 	}
 
-	start_thread(thread, stack_size);
+	queue_enqueue(&thread->queue, &thread_options.to_run, NULL);
+
+	start_thread(thread, thread_options.stack_size);
 
 	PERL_SET_CONTEXT(self);
 }
